@@ -71,10 +71,10 @@ import datetime
 import json
 import re
 from abc import ABC
-from collections import defaultdict
+from collections import OrderedDict, defaultdict
 from email.utils import parsedate_to_datetime
 from typing import Any, DefaultDict, Dict, List, Optional, Pattern, Tuple, Union
-from urllib.parse import parse_qs
+from urllib.parse import parse_qs, unquote
 from xml.etree import ElementTree as ETree
 
 import dateutil.parser
@@ -166,13 +166,7 @@ class RequestParser(abc.ABC):
         if location is not None:
             if location == "header":
                 header_name = shape.serialization.get("name")
-                if header_name in request.headers:
-                    payload = self._parse_shape(
-                        request, shape, request.headers[header_name], path_regex
-                    )
-                else:
-                    # if header is optional and not set
-                    payload = None
+                payload = request.headers.get(header_name)
             elif location == "headers":
                 payload = self._parse_header_map(shape, request.headers)
             elif location == "querystring":
@@ -181,14 +175,15 @@ class RequestParser(abc.ABC):
                 # If the query param is in the parsed dict, take the first element of the value
                 # (the parsed query params are always lists, even if only a single value).
                 payload = (
-                    next(iter(parsed_query[query_name]), None)
-                    if query_name in parsed_query
-                    else None
+                    list(iter(parsed_query[query_name])) if query_name in parsed_query else None
                 )
+                # The result is always a list. If the shape is not of type list, we extract the only element
+                if shape.type_name != "list":
+                    payload = payload[0] if payload else None
             elif location == "uri":
                 regex_group_name = shape.serialization.get("name")
                 match = path_regex.match(request.path)
-                payload = match.group(regex_group_name)
+                payload = unquote(match.group(regex_group_name))
             else:
                 raise RequestParserError("Unknown shape location '%s'." % location)
         else:
@@ -294,13 +289,28 @@ class RequestParser(abc.ABC):
         """
         if request_uri is None:
             return None
+        # Escape characters within the request_uri for the usage in a regex
+        escaped_request_uri = re.escape(request_uri)
         # Replace the variable placeholder (f.e. {ConnectionId} in "/fuu/{ConnectionId}/bar/") with a
         # named regex group. Note: allowing slashes in param values, as required, e.g., for ARNs in paths.
+        # The curly braces have been escaped previously and need to be captured here. Hence the "\\\\" in the pattern.
+        # The request URL may contain the first query parameter (starting with a ?). Hence the exclusion of ? and &.
+        # A group will have the following regex: /fuu/(?P<\g<VariableName>>[^?&]+))/bar
         request_uri_regex = re.sub(
-            "{(?P<VariableName>[^}]+)}", r"(?P<\g<VariableName>>.+)", request_uri
+            "\\\\{(?P<VariableName>[^}]+)\\\\}",
+            r"(?P<\g<VariableName>>[^?&]+)",
+            escaped_request_uri,
         )
-        # Put regex in fences to make sure that we do not match any substrings
-        request_uri_regex = f"^{request_uri_regex}$"
+        # The variable name (now regex group name) can also contain trailing "+" signs (f.e. S3).
+        # This means that the key should be greedy (there can only be one greedy param).
+        # This is implemented by making groups which do not contain a "+" and are not at the end of the URI non-greedy
+        # (with a trailing ?).
+        if "\\+>[^?&]+)" in request_uri_regex:
+            request_uri_regex = request_uri_regex.replace("\\+>[^?&]+)", ">[^?&]_greedy_+)")
+            request_uri_regex = request_uri_regex.replace(">[^?&]+)", ">[^?&]+?)")
+            request_uri_regex = request_uri_regex.replace(">[^?&]_greedy_+)", ">[^?&]+)")
+        # Make sure the URL starts with our regex, do not add a fence to the end (there might be optional query params).
+        request_uri_regex = f"^{request_uri_regex}"
         # The result is a regex itself.
         return re.compile(request_uri_regex)
 
@@ -507,32 +517,57 @@ class BaseRestRequestParser(RequestParser):
         # We create a mapping when the parser is initialized.
         # Since the path can contain URI path parameters, the key of the dict is a regex.
         self.operation_lookup: DefaultDict[
-            str, DefaultDict[Pattern[str], OperationModel]
-        ] = defaultdict(lambda: defaultdict())
-        for operation in service.operation_names:
-            operation_model = service.operation_model(operation)
+            str, OrderedDict[Pattern[str], OperationModel]
+        ] = defaultdict(lambda: OrderedDict())
+        # Extract all operation models from the service spec
+        operation_models = [
+            service.operation_model(operation) for operation in service.operation_names
+        ]
+        # Sort the operation models descending based on their normalized request URIs.
+        # This is necessary, to ensure that greedy regex matches do not cause wrong method lookups.
+        # f.e. /fuu/{bar}/baz should have precedence over /fuu/{bar}.
+        sorted_operation_models = sorted(
+            operation_models, key=self._get_normalized_request_uri_length, reverse=True
+        )
+        for operation_model in sorted_operation_models:
             http = operation_model.http
-            if len(http) > 0:
-                method = http.get("method")
-                request_uri = http.get("requestUri")
-                request_uri_regex = self._convert_request_uri_to_regex(request_uri)
-                self.operation_lookup[method][request_uri_regex] = operation_model
+            method = http.get("method")
+            request_uri_regex = self._get_request_uri_regex(operation_model)
+            self.operation_lookup[method][request_uri_regex] = operation_model
+
+    def _get_normalized_request_uri_length(self, operation_model: OperationModel) -> str:
+        """
+        Fings the normalized request URI for the given operation model.
+        A normalized request URI has a static, common replacement for path parameter placeholders, starting with a
+        space character (which is the lowest non-control character in ASCII and is not expected to be present in a
+        service specification's request URI pattern).
+        This allows the resulting normalized request URIs to be sorted.
+        :param operation_model: to get the normalized request URI for.
+            This function expects that the given operation model has HTTP metadata!
+        :return: normalized request URI for the given operation model
+        """
+        request_uri: str = operation_model.http.get("requestUri")
+        # Make sure that all path parameter placeholders have the same name and length
+        return re.sub(r"{(.*?)}", " param", request_uri)
 
     def parse(self, request: HttpRequest) -> Tuple[OperationModel, Any]:
+        # Use the path and the query string for the matching
+        url = request.path
+        if request.query_string:
+            url += f"?{to_str(request.query_string)}"
         # Find the regex which matches the given path (as well as its operation)
-        # Note: using list(filter(..)) here, to avoid "TypeError: StopIteration interacts badly with generators"
-        matching = list(
-            filter(
-                lambda item: item[0].match(request.path),
-                self.operation_lookup[request.method].items(),
+        try:
+            path_regex, operation = next(
+                filter(
+                    lambda item: item[0].match(url),
+                    self.operation_lookup[request.method].items(),
+                )
             )
-        )
-        if not matching:
+        except StopIteration:
             raise RequestParserError(
                 f"Unable to find operation for request to service "
                 f"{self.service.service_name}: {request.method} {request.path}"
             )
-        path_regex, operation = matching[0]
         shape: StructureShape = operation.input_shape
         final_parsed = {}
         if shape is not None:
@@ -548,28 +583,40 @@ class BaseRestRequestParser(RequestParser):
         final_parsed: dict,
     ) -> None:
         """Parses all attributes which are located in the payload / body of the incoming request."""
+        payload_parsed = {}
+        non_payload_parsed = {}
         if "payload" in shape.serialization:
             # If a payload is specified in the output shape, then only that shape is used for the body payload.
             payload_member_name = shape.serialization["payload"]
             body_shape = member_shapes[payload_member_name]
             if body_shape.serialization.get("eventstream"):
                 body = self._create_event_stream(request, body_shape)
-                final_parsed[payload_member_name] = body
-            elif body_shape.type_name in ["string", "blob"]:
-                # This is a stream
-                body = request.data
-                if isinstance(body, bytes):
-                    body = body.decode(self.DEFAULT_ENCODING)
-                final_parsed[payload_member_name] = body
+                payload_parsed[payload_member_name] = body
+            elif body_shape.type_name == "string":
+                # Only set the value if it's not empty (the request's data is an empty binary by default)
+                if request.data:
+                    body = request.data
+                    if isinstance(body, bytes):
+                        body = body.decode(self.DEFAULT_ENCODING)
+                    payload_parsed[payload_member_name] = body
+            elif body_shape.type_name == "blob":
+                # Only set the value if it's not empty (the request's data is an empty binary by default)
+                if request.data:
+                    payload_parsed[payload_member_name] = request.data
             else:
                 original_parsed = self._initial_body_parse(request.data)
-                final_parsed[payload_member_name] = self._parse_shape(
+                payload_parsed[payload_member_name] = self._parse_shape(
                     request, body_shape, original_parsed, path_regex
                 )
+        else:
+            # The payload covers the whole body. We only parse the body if it hasn't been handled by the payload logic.
+            non_payload_parsed = self._initial_body_parse(request.data)
         # even if the payload has been parsed, the rest of the shape needs to be processed as well
-        original_parsed = self._initial_body_parse(request.data)
-        body_parsed = self._parse_shape(request, shape, original_parsed, path_regex)
-        final_parsed.update(body_parsed)
+        # (for members which are located outside of the body, like uri or header)
+        non_payload_parsed = self._parse_shape(request, shape, non_payload_parsed, path_regex)
+        # update the final result with the parsed body and the parsed payload (where the payload has precedence)
+        final_parsed.update(non_payload_parsed)
+        final_parsed.update(payload_parsed)
 
     def _initial_body_parse(self, body_contents: bytes) -> any:
         """
@@ -615,15 +662,15 @@ class RestXMLRequestParser(BaseRestRequestParser):
         parsed = {}
         xml_dict = self._build_name_to_xml_node(node)
         for member_name, member_shape in shape.members.items():
-            if "location" in member_shape.serialization or member_shape.serialization.get(
-                "eventheader"
-            ):
-                # All members with locations have already been handled in ``_parse_non_payload_attrs``,
-                # so we don't need to parse these members.
-                continue
             xml_name = self._member_key_name(member_shape, member_name)
             member_node = xml_dict.get(xml_name)
-            if member_node is not None:
+            # If a shape defines a location trait, the node might be None (since these are extracted from the request's
+            # metadata like headers or the URI)
+            if (
+                member_node is not None
+                or "location" in member_shape.serialization
+                or member_shape.serialization.get("eventheader")
+            ):
                 parsed[member_name] = self._parse_shape(
                     request, member_shape, member_node, path_regex
                 )
@@ -813,7 +860,8 @@ class JSONRequestParser(BaseJSONRequestParser):
 
     def parse(self, request: HttpRequest) -> Tuple[OperationModel, Any]:
         target = request.headers["X-Amz-Target"]
-        _, operation_name = target.split(".")
+        # assuming that the last part of the target string (e.g., "x.y.z.MyAction") contains the operation name
+        operation_name = target.rpartition(".")[2]
         operation = self.service.operation_model(operation_name)
         shape = operation.input_shape
         path_regex = self._get_request_uri_regex(operation)
